# -*- python -*-
# $RCSfile: scriptloader.py,v $
# $Revision: 1.10 $
# $Author: reida $
# $Date: 2005/07/22 13:37:56 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# ScriptLoader is a derived class of InteractiveInterpreter that
# provides a correct traceback when exceptions occur while processing
# a Python file.  In addition, it provides a hook that can be
# redefined in a derived class to update a progress bar.

# The problem with InteractiveInterpreter is that when an exception is
# raised, the line numbers in the traceback are offsets into the
# current code block, and the current code block is not the file being
# processed.  It is instead just the set of lines passed to runsource
# since the last successful execution.  The traceback uses these
# erroneous line numbers to find the source code for the offending
# instructions.  The filename passed to runsource is only used in
# error messages and to find the source code.  The solution presented
# here is to pass a fake filename to runsource.  The fake name encodes
# the line number of the beginning of the code block.  Then when an
# exception occurs, the actual line numbers can be computed, and the
# traceback can be corrected.

# Because it's difficult to actually create and modify a traceback
# object, this code only creates a modified traceback list, as
# returned by traceback.extract_tb.  That means that the modified
# traceback can't be passed to sys.excepthook.  Instead we pass the
# traceback list to excepthook.printTraceBack().

# Portions of this file have been copied directly from code.py in the
# Python Library.

import code, sys, traceback
from oof2.common import excepthook
from oof2.common import debug

magicstring = 'O#O#F'    # used to construct fake file name
lenmagic = len(magicstring)

class ScriptLoader(code.InteractiveInterpreter):
    def __init__(self, filename, locals=None):
        self.filename = filename
        # If the open is going to fail, we really want this to
        # happen as early as possible.
        try:
            self.fileobj = open(self.filename)
        except IOError:
            raise
        code.InteractiveInterpreter.__init__(self, locals)
        self.resetbuffer()
        self.startblock = 0             # line number of a code block start
        self.currentline = 0
        self.error = 0

    def resetbuffer(self):
        self.buffer = []

    def push(self, line, lineno):
        self.buffer.append(line)
        source = "\n".join(self.buffer)
        more = self.runsource(source, magicstring+str(self.startblock))
        if not more:
            self.resetbuffer()
            self.startblock = lineno + 1 # start of next code block
        return more

    def run(self):
        # file = open(self.filename)
        debug.fmsg("ScriptLoader...")
        self.lines = self.fileobj.readlines()
        nlines = len(self.lines)
        for i in range(nlines):
            line = self.lines[i]
            self.currentline = i+1
            more = self.push(line[:-1], i)        # omit trailing newline
            self.progress(i+1, nlines)
            if self.stop() or self.error:
                break
        # If the file ended inside an indented block, the block won't
        # have been executed yet.  push()ing an empty line forces the
        # block to close.
        if more:
            self.push("", nlines)
        self.done()

    def showtraceback(self):
        self.error = 1
        (type, value, tb) = sys.exc_info()
        tblist = traceback.extract_tb(tb)
        # python docs make obscure comments about avoiding circular
        # references to tracebacks.
        tb = None

        fixedtb = []
        for frame in tblist[1:]:  # frame = (filename, lineno, func, text)
            if frame[0][:lenmagic] == magicstring:
                lineno = int(frame[0][lenmagic:]) + frame[1]
                fixedtb.append((self.filename, lineno, frame[2],
                                self.lines[lineno-1][:-1]))
            else:                 # this frame is not from the file
                fixedtb.append((frame[0], frame[1], frame[2], frame[3]))
        excepthook.printTraceBack(type, value, fixedtb)

    def showsyntaxerror(self, filename=None):
        self.error = 1
        (etype, value, tb) = sys.exc_info()
        # The traceback information in tb is completely useless, but
        # fortunately it just so happens that we know what line of the
        # file we're on, so we don't need it.  All we really need is the
        # offset, which tells us where in the line the error occurred.
        if etype is SyntaxError:
            try:
                msg, (dummy_filename, lineno, offset, line) = value
            except:
                pass
            else:
                value = SyntaxError(msg, (self.filename, self.currentline,
                                          offset,
                                          self.lines[self.currentline-1]))
        excepthook.printTraceBack(etype, value, [])
             

    def progress(self, current, total):
        pass

    def stop(self):
        return 0

    def done(self):
        self.fileobj.close()
