# -*- python -*-
# $RCSfile: outputDefs.py,v $
# $Revision: 1.64 $
# $Author: langer $
# $Date: 2005/09/12 17:25:24 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file generates concrete Outputs from the Outputs defined in
# outputClones.py.  The concrete Outputs are stored in a tree
# structure which can easily be converted to a menu.

from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import field
from oof2.SWIG.engine import flux
from oof2.SWIG.engine import outputval
from oof2.SWIG.engine import planarity
from oof2.SWIG.engine.IO import propertyoutput
from oof2.common import debug
from oof2.common import enum
from oof2.common import primitives
from oof2.common.IO import output
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import problem
from oof2.engine.IO import outputClones

from types import *
import string

######################################

# Displacement

# Although Displacement is a Field, it's convenient to treat it as a
# Point so that it can be added to a position.  Defining
# DisplacementOutput in terms of another output which sets a parameter
# to problem.Displacement allows the computability checking to work.

Displacement = field.getCompoundField("Displacement")
iter = Displacement.iterator(planarity.IN_PLANE)
disp0 = iter.cloneIndex()
iter.next()
disp1 = iter.cloneIndex()

def _disp2point(fieldlist, *args, **kwargs):
    # Convert displacement field OutputVals to Points
    return [primitives.Point(field[disp0], field[disp1]) for field in fieldlist]

displacementFieldOutput = outputClones.FieldAnywayOutput.clone(
    name="displacement field")
displacementFieldOutput.plugParameter(which="field", value=Displacement)
displacementOutput = output.Output(name="dispPoint",
                                   oType=primitives.Point,
                                   func=_disp2point,
                                   inputs=output.Input("displacement",
                                                       outputval.OutputValPtr,
                                                       displacementFieldOutput))

# Enhanced position is scalefactor*Displacement + original position
enhancedPosition = outputClones.PointSumOutput.clone(
    name='enhanced position',
    tip='Exaggerated displacement field.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/output/enhancedPosOutput.xml'))
enhancedPosition.plugInput(displacementOutput, 'point1')
enhancedPosition.plugInput(outputClones.posOutput, 'point2')
enhancedPosition.plugParameter(value=1.0, which='b')
enhancedPosition.plugParameter(default=1.0, which='a', rename='factor',
                               tip='Displacement multiplier.')

# Actual position is enhanced position with a scalefactor of 1
actualPosition = enhancedPosition.clone(
    name='actual position',
    tip='Displaced position.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/output/actualPosOutput.xml')
    )
actualPosition.plugParameter(value=1.0, which='factor')

originalPosition = outputClones.posOutput.clone(
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/output/originalPosOutput.xml'))

output.definePositionOutput('original', originalPosition)
output.definePositionOutput('actual', actualPosition)
output.definePositionOutput('enhanced', enhancedPosition)

######################

output.defineAggregateOutput('Field:Value', outputClones.FieldOutput)
output.defineAggregateOutput('Field:Derivative', outputClones.FieldDerivOutput)
output.defineAggregateOutput('Field:Invariant',
                             outputClones.FieldInvariantOutput)
output.defineAggregateOutput('Flux:Value', outputClones.FluxOutput)
output.defineAggregateOutput('Flux:Invariant', outputClones.FluxInvariantOutput)

output.defineScalarOutput('Field:Component', outputClones.FieldCompOutput)
output.defineScalarOutput('Field:Invariant', outputClones.FieldInvariantOutput)
output.defineScalarOutput('Field:Derivative:Component',
                          outputClones.FieldDerivCompOutput)
output.defineScalarOutput('Field:Derivative:Invariant',
                          outputClones.FieldDerivInvariantOutput)
output.defineScalarOutput('Flux:Component', outputClones.FluxCompOutput)
output.defineScalarOutput('Flux:Invariant', outputClones.FluxInvariantOutput)

from oof2.common import strfunction
xyfunc = outputClones.FunctionOutput.clone(
    tip='Compute an arbitrary function of x and y.',
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/output/funcOutput.xml'))
xyfunc.plugParameter(default=strfunction.XYStrFunction('1.0'), which='f')
output.defineScalarOutput('XYFunction', xyfunc)

###########

negElectricFieldOutput = outputClones.FieldDerivOutput.clone()
negElectricFieldOutput.plugParameter(which="field",
                                     value=field.getCompoundField("Voltage"))
electricFieldOutput = outputClones.NegateOutput.clone()
electricFieldOutput.plugInput(negElectricFieldOutput, "scalee")
#output.defineAggregateOutput('E Field', electricFieldOutput)
eFieldCompOutput = outputClones.ComponentOutput.clone()
eFieldCompOutput.plugInput(electricFieldOutput, 'field')
#output.defineScalarOutput('E Field:component', eFieldCompOutput)

## TODO: Uncomment the above two defineOutput lines, but only after
## fixing the problms with the MeshParamWidgets.  The problem is that
## the widget for the component of the E field looks for a
## FieldWidget, and doesn't find one because the Field isn't a
## settable parameter.


######################

# PropertyOutputs, which are computed by the Property objects in a
# Material, only have to be identified by name, type, and (possibly)
# initializer here.  The computations are all done by the
# PropertyOutput<TYPE> classes and the various Property::output virtual
# functions.

class EnergyType(enum.EnumClass(
    ("Total", "All contributions to the energy"),
    ("Elastic", "Elastic energy-- one half stress times geometric strain"),
    ("Electric", "Electric energy-- one half total polarization times electric field")
    )):
    tip="Varieties of Energy."
    discussion="""<para>
    <classname>EnergyType</classname> indicates which energy is to be
    computed by the <xref linkend='Output:Energy'/> Output.
    </para>"""

propertyoutput.ScalarPropertyOutputRegistration(
    "Energy",
    parameters=enum.EnumParameter("etype", EnergyType, default="Total",
                                  tip='The type of energy to compute.'),
    tip='Compute an energy density.',
    discussion="""
      <para>Compute the energy density of the &fields; on the &mesh;.
      Different values of <varname>etype</varname> include different
      contributions to the energy.</para>"""
    )


    
